from django.urls import path, include

from . import views


app_name = "htmx"
urlpatterns = [
    path("", views.incident_list, name="incident-list"),
    path("<int:pk>/", views.incident_detail, name="incident-detail"),
    path("<int:pk>/ack/", views.incident_add_ack, name="incident-add-ack"),
    path("<int:pk>/ack-detail/", views.incident_detail_add_ack, name="incident-detail-add-ack"),
    path("<int:pk>/close-detail/", views.incident_detail_close, name="incident-detail-close"),
    path("<int:pk>/reopen-detail/", views.incident_detail_reopen, name="incident-detail-reopen"),
    path("<int:pk>/ticket-edit-detail/", views.incident_detail_edit_ticket, name="incident-detail-edit-ticket"),
    path("<int:pk>/ticket-add-detail/", views.incident_detail_add_ticket, name="incident-detail-add-ticket"),
]



from django.urls import path

from . import views


app_name = "htmx"
urlpatterns = [
    path("names/", views.dateformat_names, name="dateformat-names"),
    path("change/", views.change_dateformat, name="change-dateformat"),
]



from __future__ import annotations

import logging
from typing import Optional
from datetime import datetime

from django.conf import settings
from django.contrib.auth.models import Group
from django.contrib.auth import get_user_model
from django.contrib import messages
from django.core.exceptions import PermissionDenied
from django.shortcuts import redirect, render, get_object_or_404
from django.urls import reverse

from django.views.decorators.http import require_GET, require_POST
from django.core.paginator import Paginator
from django.http import HttpRequest, HttpResponse
from django_htmx.middleware import HtmxDetails

from argus.incident.models import Incident
from argus.util.datetime_utils import make_aware

from .customization import get_incident_table_columns
from .utils import get_filter_function
from .forms import AckForm, DescriptionOptionalForm, EditTicketUrlForm, AddTicketUrlForm


User = get_user_model()
LOG = logging.getLogger(__name__)
DEFAULT_PAGE_SIZE = getattr(settings, "ARGUS_INCIDENTS_DEFAULT_PAGE_SIZE", 10)
ALLOWED_PAGE_SIZES = getattr(settings, "ARGUS_INCIDENTS_PAGE_SIZES", [10, 20, 50, 100])


def prefetch_incident_daughters():
    return Incident.objects.select_related("source").prefetch_related(
        "incident_tag_relations",
        "incident_tag_relations__tag",
        "events",
        "events__ack",
    )


class HtmxHttpRequest(HttpRequest):
    htmx: HtmxDetails


# fetch with htmx
def incident_row(request, pk: int):
    incident = get_object_or_404(Incident, d=pk)
    context = {
        "incident": incident,
    }
    return render(request, "htmx/incidents/_incident_row.html", context=context)


def incident_detail(request, pk: int):
    incident = get_object_or_404(Incident, id=pk)
    action_endpoints = {
        "ack": reverse("htmx:incident-detail-add-ack", kwargs={"pk": pk}),
        "close": reverse("htmx:incident-detail-close", kwargs={"pk": pk}),
        "reopen": reverse("htmx:incident-detail-reopen", kwargs={"pk": pk}),
        "edit_ticket": reverse("htmx:incident-detail-edit-ticket", kwargs={"pk": pk}),
        "add_ticket": reverse("htmx:incident-detail-add-ticket", kwargs={"pk": pk}),
    }
    context = {
        "incident": incident,
        "endpoints": action_endpoints,
        "page_title": str(incident),
    }
    return render(request, "htmx/incidents/incident_detail.html", context=context)


def _incident_add_ack(pk: int, formdata, user: User, group: Optional[str] = None):
    incident = get_object_or_404(Incident, id=pk)
    is_group_member = None
    if group:
        group = get_object_or_404(Group, name=group)
        is_group_member = user.groups.filter(pk=group.pk).exists()
    form = AckForm()
    if formdata:
        if group and not is_group_member:
            raise PermissionDenied("User {request.user} is not a member of the correct group")
        form = AckForm(formdata)
        if form.is_valid():
            incident.create_ack(
                user,
                description=form.cleaned_data["description"],
                expiration=form.cleaned_data["expiration"],
            )
    context = {
        "form": form,
        "incident": incident,
        "page_title": str(incident),
        "group": group,
        "is_group_member": is_group_member,
    }
    return incident, context


def incident_add_ack(request, pk: int, group: Optional[str] = None):
    formdata = request.POST or None
    _, context = _incident_add_ack(pk, formdata, request.user, group)
    return render(request, "htmx/incidents/incident_add_ack.html", context=context)


@require_POST
def incident_detail_add_ack(request, pk: int, group: Optional[str] = None):
    formdata = request.POST or None
    _incident_add_ack(pk, formdata, request.user, group)
    return redirect("htmx:incident-detail", pk=pk)


@require_POST
def incident_detail_close(request, pk: int):
    incident = get_object_or_404(Incident, id=pk)
    if not incident.stateful:
        LOG.warning(f"Attempt at closing the uncloseable {incident}")
        messages.warning(request, f"Did not close {incident}, stateless incidents cannot be closed.")
        return redirect("htmx:incident-detail", pk=pk)
    form = DescriptionOptionalForm(request.POST or None)
    if form.is_valid():
        incident.set_closed(
            request.user,
            description=form.cleaned_data.get("description", ""),
        )
        LOG.info(f"{{ incident }} manually closed by {{ request.user }}")
    return redirect("htmx:incident-detail", pk=pk)


@require_POST
def incident_detail_reopen(request, pk: int):
    incident = get_object_or_404(Incident, id=pk)
    if not incident.stateful:
        LOG.warning(f"Attempt at reopening the unopenable {incident}")
        messages.warning(request, f"Did not reopen {incident}, stateless incidents cannot be reopened.")
        return redirect("htmx:incident-detail", pk=pk)
    form = DescriptionOptionalForm(request.POST or None)
    if form.is_valid():
        incident.set_open(
            request.user,
            description=form.cleaned_data.get("description", ""),
        )
        LOG.info(f"{{ incident }} manually reopened by {{ request.user }}")
    return redirect("htmx:incident-detail", pk=pk)


@require_POST
def incident_detail_add_ticket(request, pk: int):
    incident = get_object_or_404(Incident, id=pk)
    form = AddTicketUrlForm()
    if request.POST:
        form = AddTicketUrlForm(request.POST)
        if form.is_valid():
            incident.ticket_url = form.cleaned_data["ticket_url"]
            incident.save()

    return redirect("htmx:incident-detail", pk=pk)


@require_POST
def incident_detail_edit_ticket(request, pk: int):
    incident = get_object_or_404(Incident, id=pk)
    form = EditTicketUrlForm()
    if request.POST:
        form = EditTicketUrlForm(request.POST)
        if form.is_valid():
            incident.ticket_url = form.cleaned_data["ticket_url"]
            incident.save()

    return redirect("htmx:incident-detail", pk=pk)


def _get_page_size(params):
    try:
        if (page_size := int(params.pop("page_size", DEFAULT_PAGE_SIZE))) in ALLOWED_PAGE_SIZES:
            return page_size
    except ValueError:
        pass
    return DEFAULT_PAGE_SIZE


def incident_list(request: HtmxHttpRequest) -> HttpResponse:
    columns = get_incident_table_columns()

    # Load incidents
    qs = prefetch_incident_daughters().order_by("-start_time")
    total_count = qs.count()
    last_refreshed = make_aware(datetime.now())

    params = dict(request.GET.items())

    incident_list_filter = get_filter_function()
    filter_form, qs = incident_list_filter(request, qs)
    filtered_count = qs.count()

    # Standard Django pagination
    page_num = params.pop("page", "1")
    page_size = _get_page_size(params)
    paginator = Paginator(object_list=qs, per_page=page_size)
    page = paginator.get_page(page_num)

    # The htmx magic - use a different, minimal base template for htmx
    # requests, allowing us to skip rendering the unchanging parts of the
    # template.
    if request.htmx:
        base_template = "htmx/incidents/responses/_incidents_table_refresh.html"
    else:
        base_template = "htmx/incidents/_base.html"
    context = {
        "columns": columns,
        "filtered_count": filtered_count,
        "count": total_count,
        "filter_form": filter_form,
        "page_title": "Incidents",
        "base": base_template,
        "page": page,
        "last_refreshed": last_refreshed,
        "update_interval": 30,
        "page_size": page_size,
        "all_page_sizes": ALLOWED_PAGE_SIZES,
    }

    return render(request, "htmx/incidents/incident_list.html", context=context)



"""
How to use:

Append the "context_processors" list for the TEMPLATES-backend
``django.template.backends.django.DjangoTemplates`` with the full dotted path.

See django settings for ``TEMPLATES``.
"""

from .dateformat.constants import DATETIME_DEFAULT, DATETIME_FORMATS


def theme_via_GET(request):
    # Move to theme-app?
    theme = request.GET.get("theme", None)
    return {"theme": theme}


def theme_via_session(request):
    # Move to theme-app?
    theme = request.session.get("theme", None)
    return {"theme": theme}


def datetime_format_via_session(request):
    datetime_format_name = request.session.get("datetime_format_name")
    if datetime_format_name not in DATETIME_FORMATS:
        datetime_format_name = DATETIME_DEFAULT
    # The named format always wins
    datetime_format = DATETIME_FORMATS[datetime_format_name]
    return {
        "datetime_format": datetime_format,
        "datetime_format_name": datetime_format_name,
    }



import importlib

from django.conf import settings

FUNCTION_NAME = "incident_list_filter"
DEFAULT_MODULE = "argus_htmx.incidents.filter"


def get_filter_function():
    dotted_path = getattr(settings, "ARGUS_HTMX_FILTER_FUNCTION", DEFAULT_MODULE)
    module = importlib.import_module(dotted_path)
    function = getattr(module, FUNCTION_NAME, None)
    if function:
        return function
    raise ImportError(f"Could not import {FUNCTION_NAME} from {dotted_path}")



import pathlib
from django.apps import AppConfig


class HtmxFrontendConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    label = "argus_htmx"
    name = "argus_htmx"

    def tailwind_css_files(self):
        yield from pathlib.Path(__file__).parent.glob("tailwindtheme/snippets/*.css")



from argus.site.settings._serializers import ListAppSetting


__all__ = ["APP_SETTINGS"]

_app_settings = [
    {
        "app_name": "argus_htmx",
        "urls": {
            "path": "",
            "urlpatterns_module": "argus_htmx.urls",
        },
        "context_processors": [
            "argus_htmx.context_processors.theme_via_session",
            "argus_htmx.context_processors.datetime_format_via_session",
        ],
        "middleware": {
            "argus_htmx.middleware.LoginRequiredMiddleware": "end",
        },
    },
    {
        "app_name": "django_htmx",
        "middleware": {
            "django_htmx.middleware.HtmxMiddleware": "end",
        },
    },
    {
        "app_name": "widget_tweaks",  # indent for readability
    },
]

APP_SETTINGS = ListAppSetting(_app_settings).root

LOGIN_URL = "/accounts/login/"
LOGOUT_URL = "/accounts/logout/"
LOGIN_REDIRECT_URL = "/incidents/"
LOGOUT_REDIRECT_URL = "/"

PUBLIC_URLS = [
    "/accounts/login/",
    "/api/",
    "/oidc/",
]



import json
import pathlib
import textwrap
from django.apps import apps
from django.core.management.base import BaseCommand
from django.conf import settings
from django.template import engines
from django.template.context import make_context
from django.template.loader import get_template

from argus_htmx.themes.utils import get_themes
from argus_htmx import settings as argus_htmx_settings


# Copied from https://github.com/GEANT/geant-argus/pull/15 with minor modifications
class Command(BaseCommand):
    help = """
    Uses the template specified in the TAILWIND_CONFIG_TEMPLATE setting
    (default: tailwind/tailwind.config.js) to dynamically build a tailwind.config.js.
    This file must be in an app's templates directory. The template may contain:

     - a '{{ projectpaths }} section without square brackets that will be
        popuplated  with auto discovered template dirs of installed apps
     - a '{{ daisyuithemes }}' section without square brackets that will be
        popuplated by the daisyUI theme list specified in the DAISYUI_THEMES
        setting (default: ["dark", "light", {"argus"": {...}} ])
     - a '{{ themeoverride }}' section that will be popuplated by a dict containing
        tailwind theme options specified in TAILWIND_THEME_OVERRIDE setting
        (default: {})

    This command also generates a `styles.css` that contains the base css file for
    tailwind to create its final css file. It uses the template specified by the
    `TAILWIND_CSS_TEMPLATE` setting (default: tailwind/styles.css). This file should
    also be in an app's templates directory

    The `styles.css` template may iterate over a `cssfiles` context variable that
    contains css files/snippets that should be included in the final css file. Apps
    may define a `tailwind_css_files` method that gives the location (as a str or
    pathlib.Path object) for every css file from this app to include. (see
    argus_htmx.apps.HtmxFrontendConfig for an example)

    Additional settings that govern the functionality of this command are:

     - TAILWIND_CONFIG_TARGET: the target location for writing the tailwind.config.js
     - TAILWIND_CSS_TARGET: override the base tailwind css file location

    These target locations are relative to the working directory where this command is
    executed
    """
    DEFAULT_CONFIG_TEMPLATE_NAME = "tailwind/tailwind.config.js"
    DEFAULT_CSS_TEMPLATE_NAME = "tailwind/styles.css"
    DEFAULT_CONFIG_TARGET = "src/argus_htmx/tailwindtheme/tailwind.config.js"
    DEFAULT_CSS_TARGET = "src/argus_htmx/tailwindtheme/styles.css"

    def handle(self, *args, **options):
        config_template_name = getattr(settings, "TAILWIND_CONFIG_TEMPLATE", self.DEFAULT_CONFIG_TEMPLATE_NAME)
        config_target_path = pathlib.Path(getattr(settings, "TAILWIND_CONFIG_TARGET", self.DEFAULT_CONFIG_TARGET))
        css_template_name = getattr(settings, "TAILWIND_CSS_TEMPLATE", self.DEFAULT_CSS_TEMPLATE_NAME)
        css_target_path = pathlib.Path(getattr(settings, "TAILWIND_CSS_TARGET", self.DEFAULT_CSS_TARGET))
        self.write_file(
            config_template_name,
            config_target_path,
            context=self.get_context(target_dir=config_target_path.parent),
            name="tailwind config",
        )
        self.write_file(
            css_template_name,
            css_target_path,
            context=self.get_context(target_dir=css_target_path.parent),
            name="tailwind base css",
        )

    def get_context(self, target_dir: pathlib.Path):
        return {
            "themeoverride": getattr(
                settings,
                "TAILWIND_THEME_OVERRIDE",
                argus_htmx_settings.TAILWIND_THEME_OVERRIDE,
            ),
            "daisyuithemes": textwrap.indent(
                json.dumps(get_themes(), indent=2),
                prefix=10 * " ",
                predicate=lambda line: line != "[\n",  # this is kinda hacky, but eh
            ),
            "projectpaths": "\n".join(f"        '{d}/**/*.html'," for d in self.get_template_dirs()),
            "cssfiles": self.get_css_files(target_dir),
        }

    def write_file(self, template_name, target_path, context, name):
        pathlib.Path(target_path).write_text(self.render(template_name=template_name, context=context))

        self.stdout.write(f"Wrote {name} to '{target_path}'")

    @staticmethod
    def render(template_name: str, context):
        template = get_template(template_name)
        return template.template.render(make_context(context, autoescape=False))

    @staticmethod
    def get_template_dirs():
        for engine in engines.all():
            yield from getattr(engine, "template_dirs", [])

    @classmethod
    def get_css_files(cls, target_dir: pathlib.Path):
        css_files = sorted(cls._iter_css_files(), key=lambda p: p.stem)
        return (cls.make_relative(p, target_dir) for p in css_files)

    @staticmethod
    def _iter_css_files():
        for app in apps.get_app_configs():
            if callable(css_files := getattr(app, "tailwind_css_files", None)):
                yield from (pathlib.Path(p) for p in css_files())

    @staticmethod
    def make_relative(path: pathlib.Path, base_path: pathlib.Path):
        try:
            return path.relative_to(base_path.absolute())
        except ValueError:
            return path



from django import forms

from argus.filter import get_filter_backend
from argus.incident.constants import Level
from argus.incident.models import SourceSystem


filter_backend = get_filter_backend()
QuerySetFilter = filter_backend.QuerySetFilter


class IncidentFilterForm(forms.Form):
    open = forms.BooleanField(required=False)
    closed = forms.BooleanField(required=False)
    acked = forms.BooleanField(required=False)
    unacked = forms.BooleanField(required=False)
    source = forms.MultipleChoiceField(
        choices=tuple(SourceSystem.objects.values_list("id", "name")),
        required=False,
    )
    maxlevel = forms.IntegerField(
        widget=forms.NumberInput(
            attrs={"type": "range", "step": "1", "min": min(Level).value, "max": max(Level).value}
        ),
        label="Level <=",
        initial=max(Level).value,
        required=False,
    )

    def _tristate(self, onkey, offkey):
        on = self.cleaned_data.get(onkey, None)
        off = self.cleaned_data.get(offkey, None)
        if on == off:
            return None
        if on and not off:
            return True
        if off and not on:
            return False

    def to_filterblob(self):
        if not self.is_valid():
            return {}

        filterblob = {}

        open = self._tristate("open", "closed")
        if open is not None:
            filterblob["open"] = open

        acked = self._tristate("acked", "unacked")
        if acked is not None:
            filterblob["acked"] = acked

        source = self.cleaned_data.get("source", [])
        if source:
            filterblob["sourceSystemIds"] = source

        maxlevel = self.cleaned_data.get("maxlevel", 0)
        if maxlevel:
            filterblob["maxlevel"] = maxlevel

        return filterblob


def incident_list_filter(request, qs):
    # TODO: initialize with chosen Filter.filter if any
    form = IncidentFilterForm(request.GET or None)

    if form.is_valid():
        filterblob = form.to_filterblob()
        qs = QuerySetFilter.filtered_incidents(filterblob, qs)
    return form, qs



from django.http import HttpResponse
from django.template import Template, RequestContext
from django.urls import path


def placeholder(request):
    template = Template(
        """{% extends "htmx/base.html" %}
        {% block main %}
        <h1>DESTINATION PLACEHOLDER</h1>
        {% endblock main %}
        """
    )
    context = RequestContext(request)
    return HttpResponse(template.render(context))


app_name = "htmx"
urlpatterns = [
    path("", placeholder, name="destination-placeholder"),
]



from argus.site.settings import get_json_env, validate_app_setting
from django.conf import settings
from argus_htmx import settings as argus_htmx_settings


def get_themes():
    return getattr(settings, "DAISYUI_THEMES", argus_htmx_settings.DAISYUI_THEMES)


def get_theme_names():
    themes = get_themes()
    theme_names = []
    for theme in themes:
        if isinstance(theme, str):
            theme_names.append(theme)
        elif isinstance(theme, dict):
            theme_names.extend(theme.keys())
    return theme_names



from django.urls import path

from . import views


app_name = "htmx"
urlpatterns = [
    path("", views.ThemeListView.as_view(), name="theme-list"),
    path("names/", views.theme_names, name="theme-names"),
    path("change/", views.change_theme, name="change-theme"),
]



import importlib_resources
import logging
from pathlib import Path
from typing import Optional

from django.conf import settings
from django.contrib import messages
from django.contrib.auth.models import Group
from django.core.exceptions import PermissionDenied
from django.shortcuts import render, reverse, get_object_or_404
from django.views.generic import ListView

from django.views.decorators.http import require_GET, require_POST
from django.http import HttpResponse, HttpResponseRedirect
from django_htmx.http import HttpResponseClientRefresh

from argus_htmx.themes.utils import get_theme_names
from argus_htmx.incidents.views import HtmxHttpRequest

LOG = logging.getLogger(__name__)
THEMES_MODULE = "argus_htmx"


class ThemeListView(ListView):
    http_method_names = ["get", "post", "head", "options", "trace"]
    template_name = "htmx/themes/themes_list.html"

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.themes = sorted(get_theme_names())

    def get_queryset(self):
        return self.themes

    def post(self, request, *args, **kwargs):
        theme = request.POST.get("theme")
        if theme in self.themes:
            request.session["theme"] = theme
            messages.success(request, f'Switched theme to "{theme}"')
        else:
            messages.warning(request, f'Failed to switch to theme "{theme}", no such theme installed')
        return HttpResponseRedirect("")


@require_GET
def theme_names(request: HtmxHttpRequest) -> HttpResponse:
    themes = sorted(get_theme_names())
    return render(request, "htmx/themes/theme_list.html", {"theme_list": themes})


@require_POST
def change_theme(request: HtmxHttpRequest) -> HttpResponse:
    themes = get_theme_names()
    theme = request.POST.get("theme")
    if theme in themes:
        request.session["theme"] = theme
        messages.success(request, f'Switched theme to "{theme}"')
        return HttpResponse(f"{theme}")
    return HttpResponseClientRefresh()



from django.conf import settings

__all__ = [
    "DATETIME_FALLBACK",
    "DATETIME_FORMATS",
    "DATETIME_DEFAULT",
]


DATETIME_FALLBACK = "LOCALE"
DATETIME_FORMATS = {
    "LOCALE": "DATETIME_FORMAT",  # default
    "ISO": "Y-m-d H:i:s",
    "RFC5322": "r",
    "EPOCH": "U",
}

_datetime_setting = getattr(settings, "ARGUS_FRONTEND_DATETIME_FORMAT", DATETIME_FALLBACK)

DATETIME_DEFAULT = DATETIME_FALLBACK
if _datetime_setting in DATETIME_FORMATS:
    DATETIME_DEFAULT = _datetime_setting



from django import forms


class AckForm(forms.Form):
    description = forms.CharField()
    expiration = forms.DateTimeField(required=False)


class DescriptionOptionalForm(forms.Form):
    "For closing/reopening"

    description = forms.CharField(required=False)


class EditTicketUrlForm(forms.Form):
    ticket_url = forms.URLField(required=False)


class AddTicketUrlForm(forms.Form):
    ticket_url = forms.URLField(required=True)



import logging

from django.contrib import messages
from django.shortcuts import render

from django.views.decorators.http import require_GET, require_POST
from django.http import HttpResponse
from django_htmx.http import HttpResponseClientRefresh

from argus_htmx.incidents.views import HtmxHttpRequest
from .constants import DATETIME_DEFAULT, DATETIME_FORMATS

LOG = logging.getLogger(__name__)


@require_GET
def dateformat_names(request: HtmxHttpRequest) -> HttpResponse:
    datetime_formats = DATETIME_FORMATS.keys()
    return render(request, "htmx/dateformat/_dateformat_list.html", {"datetime_formats": datetime_formats})


@require_POST
def change_dateformat(request: HtmxHttpRequest) -> HttpResponse:
    datetime_format_name = request.POST.get("dateformat", DATETIME_DEFAULT)
    default_format = DATETIME_FORMATS[DATETIME_DEFAULT]
    datetime_format = DATETIME_FORMATS.get(datetime_format_name, default_format)
    request.session["datetime_format"] = datetime_format
    request.session["datetime_format_name"] = datetime_format_name
    messages.success(request, f'Switched dateformat to "{datetime_format_name}"')
    return HttpResponseClientRefresh()
